/*
*
* VU Visualization 2
*
* Author: Dominik Bauer, Pascal Plank
* SS 2016
* All rights reserved.
*
*/

#include "VideoManager.hpp"
#include <iomanip>
#include <ctime>

#include <string> //TODO remove

cv::VideoCapture capture;

VideoManager::VideoManager()
	: rate(0)
	, fnumber(0)
	, length(0)
	, stopped(true)
	, modify(false)
	, curPos(0)
	, curIndex(0)
	, digits(0)
	, extension(".avi")
	, levels(4)
	, curLevel(0)
	, fl(0.4f)
	, fh(2.0f)
	, delta(0)
	, exaggeration_factor(2.0f)
	, lambda(0)
	, alpha(5.0f)
	, lambda_c(80)
	, chromAttenuation(0.1f)
	, useMask(false) {
	
}


double VideoManager::getCurrentFrame(){
	//return curPos;
	return capture.get(CV_CAP_PROP_POS_FRAMES);
}

double VideoManager::getNumberOfFrames(){
	//return length;
	return capture.get(CV_CAP_PROP_FRAME_COUNT);
}

void VideoManager::setMask(float x1, float y1, float x2, float y2, bool masked){
	if (masked){
		mask = cv::Mat(getFrameSize(), CV_8UC3, cv::Scalar(255,255,255));
		int width = getFrameSize().width;
		int height = getFrameSize().height;

		if (x1 > 1.f) x1 = 1.f;
		if (y1 > 1.f) y1 = 1.f;
		if (x2 > 1.f) x2 = 1.f;
		if (y2 > 1.f) y2 = 1.f;

		cv::Rect rect = cv::Rect(cv::Point2d(x1*width, y1*height), cv::Point2d(x2*width, y2*height));
		cv::rectangle(mask, rect, cv::Scalar(0,0,0), -1);
	}
	useMask = masked;
}


// ######################################################
// #################### Input/Output ####################
// ######################################################
/**
* This method sets the filename and opens a new videofile, already open videos are closed
*
* @param fileName the filename of the video to be managed
*
* @return true if new videofile could be opened
*/
bool VideoManager::setInput(const std::string &fileName)
{
	fnumber = 0;

	// Dismiss old video
	if (isOpened()){
		capture.release();
	}

	// Open new video file
	if (capture.open(fileName)){
		length = (long)capture.get(CV_CAP_PROP_FRAME_COUNT);
		rate = capture.get(CV_CAP_PROP_FPS); //getFrameRate();

		//loadVideo();
		return true;
	}
	else {
		return false;
	}
}

/**
* This method sets the output variables for explicitly saving videofiles
*
* @param filename outputfilename
* @param codec codec to be used for saving, if 0 the inputcodec is used
* @param framerate framerate the video should be saved, if 0.0f the inputframerate is used
* @param isColor indicates whether the video should be saved as colorvideo
*
* @return true if writer could be opened
*/
bool VideoManager::setOutput(const std::string &filename, int codec, double framerate, bool isColor) {
	outputFile = filename;
	extension.clear();

	if (framerate == 0.0) {
		framerate = getFrameRate(); // same as input
	}

	char c[4];
	if (codec == 0) {
		codec = getCodec(c);
	}

	return writer.open(outputFile, codec, framerate, getFrameSize(), isColor);
}

/**
* This method saves a videofile with the parameters specified in VideoManager::setOutput
*/
void VideoManager::writeOutput() {
	cv::Mat input;

	if (!isOpened() || !writer.isOpened()) {
		return;
	}

	long pos = curPos;
	jumpTo(0);

	while (getNextFrame(input)) {
		if (outputFile.length() != 0) {
			writeNextFrame(input);
		}
	}

	modify = false;
	writer.release();
	jumpTo(pos);
}

/**
* This method writes the next frame via an open writer
*
* @param frame the frame to be written
*/
void VideoManager::writeNextFrame(cv::Mat &frame) {
	if (extension.length()) {
		std::stringstream name;
		name << outputFile << std::setfill('0') << std::setw(digits) << curIndex++ << extension;
		cv::imwrite(name.str(), frame);
	}
	else {
		writer.write(frame);
	}
}

/**
* This method creates a new temporary videofile to be worked with
*
* @param framerate framerate the video should be saved, if 0.0f the inputframerate is used
* @param isColor indicates whether the video should be saved as colorvideo
*
* @return true if writer could be opened
*/
bool VideoManager::createTemp(double framerate, bool isColor) {
	std::stringstream filename;
	auto t = std::time(nullptr);
	auto tm = *std::localtime(&t);
	filename << "../Data/vidmag_" << std::put_time(&tm, "%Y_%m_%d_%H_%M_%S") << ".avi";
	tempFile = filename.str();

	tempFileList.push_back(tempFile);

	if (framerate == 0.0) {
		framerate = getFrameRate();
	}

	// Open output video
	//return tempWriter.open(tempFile, CV_FOURCC('M', 'J', 'P', 'G'), framerate, getFrameSize(), isColor);

	// -1 for choosing codec
	//return tempWriter.open(tempFile, -1, framerate, getFrameSize(), isColor);

	// use MPEG4 as codec
	return tempWriter.open(tempFile, cv::VideoWriter::fourcc('M', 'P', '4', 'V'), framerate, getFrameSize(), isColor);
}





// #################################################################
// #################### Magnification functions ####################
// #################################################################
/**
* This method implements the motion magnification on a newly created copy of the current video
*/
void VideoManager::motionMagnification() {
	// set filter
	setSpatialFilter(LAPLACIAN);
	setTemporalFilter(IIR);

	// create a temp file
	createTemp();


	cv::Mat input;
	cv::Mat output;
	cv::Mat motion;

	std::vector<cv::Mat> pyramid;
	std::vector<cv::Mat> filtered;

	// if no capture device has been set
	if (!isOpened()) {
		return;
	}

	modify = true;
	stopped = false;
	long pos = curPos;
	jumpTo(0);

	while (!isStopped()) {

		if (!getNextFrame(input))
			break;

		input.convertTo(input, CV_32FC3, 1.0 / 255.0f);

		cv::cvtColor(input, input, CV_BGR2Lab);
		cv::Mat s = input.clone();
		spatialFilter(s, pyramid);

		// temporal filtering
		if (fnumber == 0){			// first frame
			lowpass1 = pyramid;
			lowpass2 = pyramid;
			filtered = pyramid;
		} else {
			for (int i = 0; i < levels; i++) {
				curLevel = i;
				temporalFilter(pyramid.at(i), filtered.at(i));
			}

			// amplify each spatial frequency band          
			cv::Size filterSize = filtered.at(0).size();
			int w = filterSize.width;
			int h = filterSize.height;

			delta = (float)lambda_c / 8.0f / (1.0f + alpha);
			exaggeration_factor = 2.0;

			// calc the representative wavelength
			lambda = (float)sqrt(w*w + h*h) / 3.0f;		// 3 is experimental constant

			for (int i = levels; i >= 0; i--) {
				curLevel = i;

				amplify(filtered.at(i), filtered.at(i));

				// For each pyramid level the lambda goes down by a factor of 2
				lambda /= 2.0;
			}
		}

		// reconstruct motion
		reconImgFromLaplacianPyramid(filtered, levels, motion);

		// attenuate I, Q channels
		attenuate(motion, motion);

		// combine the source frame and motion image (skip the first image)
		if (fnumber > 0) {    
			s += motion;
		}

		
		output = s.clone();

		//mask
		if (useMask) input.copyTo(output, mask);

		// convert back to rgb color space and CV_8UC3
		cv::cvtColor(output, output, CV_Lab2BGR);
		output.convertTo(output, CV_8UC3, 255.0, 1.0 / 255.0);

		tempWriter.write(output);
		fnumber++;

		//std::string msg = "Processing...";
		//std::cout << "Motionmagnification: " << floor((fnumber) * 100.f / length) << std::endl;
	}

	tempWriter.release();
	setInput(tempFile);
	jumpTo(pos);
	fnumber = 0;

	//std::cout << "\n... Motion-Magnification done" << std::endl;
}

/**
* attenuate	This method attenuates the I and Q channels
*
* @param src source image
* @param dst resutling image
*/
void VideoManager::attenuate(cv::Mat &src, cv::Mat &dst) {
	cv::Mat planes[3];
	cv::split(src, planes);
	planes[1] = planes[1] * chromAttenuation;
	planes[2] = planes[2] * chromAttenuation;
	cv::merge(planes, 3, dst);
}


/**
* This method implements the color magnification on a newly created copy of the current video
*/
void VideoManager::colorMagnification() {
	//std::cout << "\n>>> Color-Magnification" << std::endl;

	setSpatialFilter(GAUSSIAN);
	setTemporalFilter(IDEAL);

	createTemp();

	cv::Mat input;
	cv::Mat output;
	cv::Mat motion;
	cv::Mat temp;

	std::vector<cv::Mat> frames;
	std::vector<cv::Mat> downSampledFrames;
	std::vector<cv::Mat> filteredFrames;

	cv::Mat videoMat;		// concatenated down-sampled frames
	cv::Mat filtered;		// concatenated filtered images

	// if no capture device has been set
	if (!isOpened()) {
		return;
	}

	modify = true;
	stopped = false;
	long pos = curPos;
	jumpTo(0);

	// Spatial filtering
	while (getNextFrame(input) && !isStopped()) {
		input.convertTo(temp, CV_32FC3);
		frames.push_back(temp.clone());
		// spatial filtering
		std::vector<cv::Mat> pyramid;
		spatialFilter(temp.clone(), pyramid);
		downSampledFrames.push_back(pyramid.at(levels - 1)); //TODO why only last pyramid level?

		//std::cout << "Spatial Filtering: " << floor((fnumber++) * 100.0f / length) << std::endl;
	}
	if (isStopped()){
		fnumber = 0;
		return;
	}

	// Concat all frames into single large Mat, each column is a reshaped single frame
	concat(downSampledFrames, videoMat);

	// Temporal filtering
	temporalFilter(videoMat, filtered);

	// Amplify color motion
	amplify(filtered, filtered);

	// De-concat the filtered image into filtered frames
	deConcat(filtered, downSampledFrames.at(0).size(), filteredFrames);

	// Amplify each frame by adding frame image and motions, then write into video
	fnumber = 0;
	for (int i = 0; i < length - 1 && !isStopped(); ++i) {
		// up-sample the motion image        
		upsamplingFromGaussianPyramid(filteredFrames.at(i), levels, motion);
		resize(motion, motion, frames.at(i).size());
		temp = frames.at(i) + motion;

		output = temp.clone();
		double minVal, maxVal;
		minMaxLoc(output, &minVal, &maxVal); //find minimum and maximum intensities
		output.convertTo(output, CV_8UC3, 255.0 / (maxVal - minVal),
			-minVal * 255.0 / (maxVal - minVal));

		if (useMask){
			frames.at(i).convertTo(temp, CV_8UC3);
			temp.copyTo(output, mask);
		}

		tempWriter.write(output);
		//frames.at(i) = output.clone();

		//std::cout << "Amplifying: " << floor((fnumber++) * 100.0 / length) << std::endl;
	}

	tempWriter.release();
	setInput(tempFile);
	jumpTo(pos);
	fnumber = 0;

	//std::cout << "\n... Color-Magnification done" << std::endl;
}

/**
* This method amplifies the motion in a given image
*
* @param filtered	- motion image
*/
void VideoManager::amplify(const cv::Mat &src, cv::Mat &dst) {
	float currAlpha;

	switch (spatialType) {
	case LAPLACIAN:
		// calc modified alpha for this level
		currAlpha = lambda / delta / 8 - 1;
		currAlpha *= exaggeration_factor;
		if (curLevel == levels || curLevel == 0)     // dismiss the highest and lowest frequency band
			dst = src * 0;
		else
			dst = src * cv::min(alpha, currAlpha);
		break;
	case GAUSSIAN:
		dst = src * alpha;
		break;
	default:
		break;
	}
}

/**
* This method concatinates the given frames into a single "image", each column is a reshaped frame
*
* @param frames	frames of the video to be concatinated
* @param dst concatinated image
*/
void VideoManager::concat(const std::vector<cv::Mat> &frames, cv::Mat &dst) {
	cv::Size frameSize = frames.at(0).size();
	cv::Mat temp(frameSize.width*frameSize.height, length - 1, CV_32FC3);
	for (int i = 0; i < length - 1; ++i) {
		cv::Mat input = frames.at(i);
		cv::Mat reshaped = input.reshape(3, input.cols*input.rows).clone();	// reshape the frame into one column
		cv::Mat line = temp.col(i);
		reshaped.copyTo(line);	// save the reshaped frame to one column of the destinate big image
	}
	temp.copyTo(dst);
}

/**
* This method de-concatenates the image src into the original frames
*
* @param src the concatinated image to be de-concatenated
* @param framesize the framesize
* @param frames	vector of frames
*/
void VideoManager::deConcat(const cv::Mat &src, const cv::Size &frameSize, std::vector<cv::Mat> &frames) {
	for (int i = 0; i < length - 1; ++i) {
		cv::Mat line = src.col(i).clone();
		cv::Mat reshaped = line.reshape(3, frameSize.height).clone();
		frames.push_back(reshaped);
	}
}






// #################################################################
// ############################ Filters ############################
// #################################################################
/**
* This method sets the spatial filter for the magnifycation
*
* @param type the type of the spatial filter
*/
void VideoManager::setSpatialFilter(spatialFilterType type) {
	spatialType = type;
}

/**
* This method sets the temporal filter for the magnifycation
*
* @param type the type of the temporal filter
*/
void VideoManager::setTemporalFilter(temporalFilterType type) {
	temporalType = type;
}

/**
* This method calls the spatial filters according to the previously set spatialType
*
* @param src source image to be filtered
* @param dst the resulting filtered image
*/
bool VideoManager::spatialFilter(const cv::Mat &src, std::vector<cv::Mat> &pyramid) {
	switch (spatialType) {
	case LAPLACIAN:     // laplacian pyramid
		return buildLaplacianPyramid(src, levels, pyramid);
		break;
	case GAUSSIAN:      // gaussian pyramid
		return buildGaussianPyramid(src, levels, pyramid);
		break;
	default:
		return false;
		break;
	}
}

/**
* This method calls the temporal filters according to the previously set temporalType
*
* @param src source image to be filtered
* @param dst the resulting filtered image
*/
void VideoManager::temporalFilter(const cv::Mat &src, cv::Mat &dst) {
	switch (temporalType) {
	case IIR:       // IIR bandpass filter
		temporalIIRFilter(src, dst);
		break;
	case IDEAL:     // Ideal bandpass filter
		temporalIdealFilter(src, dst);
		break;
	default:
		break;
	}
	return;
}

/**
* This method implements the IIR temporal filtering of a given image (of concatenated frames)
*
* @param src the source image
* @param dst the filtered destination image
*/
void VideoManager::temporalIIRFilter(const cv::Mat &src, cv::Mat &dst) {
	cv::Mat temp1 = (1 - fh)*lowpass1[curLevel] + fh*src;
	cv::Mat temp2 = (1 - fl)*lowpass2[curLevel] + fl*src;
	lowpass1[curLevel] = temp1;
	lowpass2[curLevel] = temp2;
	dst = lowpass1[curLevel] - lowpass2[curLevel];
}

/**
* This method implements the ideal temporal filtering of a given image (of concatenated frames)
*
* @param src the source image
* @param dst the filtered destination image
*/
void VideoManager::temporalIdealFilter(const cv::Mat &src, cv::Mat &dst) {
	cv::Mat channels[3];

	// split into 3 channels
	cv::split(src, channels);

	for (int i = 0; i < 3; ++i) {

		cv::Mat current = channels[i];  // current channel
		cv::Mat tempImg;

		int width = cv::getOptimalDFTSize(current.cols);
		int height = cv::getOptimalDFTSize(current.rows);

		cv::copyMakeBorder(current, tempImg, 0, height - current.rows, 0, width - current.cols, cv::BORDER_CONSTANT, cv::Scalar::all(0));

		// do the DFT
		cv::dft(tempImg, tempImg, cv::DFT_ROWS | cv::DFT_SCALE);

		// construct the filter
		cv::Mat filter = tempImg.clone();
		createIdealBandpassFilter(filter, fl, fh, rate);

		// apply filter
		cv::mulSpectrums(tempImg, filter, tempImg, cv::DFT_ROWS);

		// do the inverse DFT on filtered image
		cv::idft(tempImg, tempImg, cv::DFT_ROWS | cv::DFT_SCALE);

		// copy back to the current channel
		tempImg(cv::Rect(0, 0, current.cols, current.rows)).copyTo(channels[i]);
	}
	// merge channels
	cv::merge(channels, 3, dst);

	// normalize the filtered image
	cv::normalize(dst, dst, 0, 1, CV_MINMAX);
}

/**
* This method creates an ideal 1D bandpass-filer
*
* @param filter the filter to be created
* @param fl the lower cut-off
* @param fh the higher cut-off
* @param rate the sampling rate, i.e. video frame rate
*/
void VideoManager::createIdealBandpassFilter(cv::Mat &filter, double fl, double fh, double rate) {
	int width = filter.cols;
	int height = filter.rows;

	fl = 2 * fl * width / rate;
	fh = 2 * fh * width / rate;

	float response;

	for (int i = 0; i < height; ++i) {
		for (int j = 0; j < width; ++j) {
			// filter response
			if (j >= fl && j <= fh)
				response = 1.0f;
			else
				response = 0.0f;
			filter.at<float>(i, j) = response;
		}
	}
}






// ##################################################
// #################### Pyramids ####################
// ##################################################
/**
* This method constructs a laplacian pyramid from a given image
*
* @param img the source image
* @param levels the levels of the pyramid to be constructed, must be >= 1
* @param pyramid the constructed pyramid
*
* @return true if constuction was successfull
*/
bool VideoManager::buildLaplacianPyramid(const cv::Mat &img, const int levels, std::vector<cv::Mat> &pyramid) {
	if (levels < 1){
		std::cerr << "Levels must be larger than 1" << std::endl;
		return false;
	}

	pyramid.clear();
	cv::Mat currentImg = img;
	for (int i = 0; i < levels; i++) {
		cv::Mat down, up;
		cv::pyrDown(currentImg, down);
		cv::pyrUp(down, up, currentImg.size());
		cv::Mat lap = currentImg - up;
		pyramid.push_back(lap);
		currentImg = down;
	}
	pyramid.push_back(currentImg);
	return true;
}

/**
* This method constructs a gaussian pyramid from a given image
*
* @param img the source image
* @param levels the levels of the pyramid to be constructed, must be >= 1
* @param pyramid the constructed pyramid
*
* @return true if constuction was successfull
*/
bool VideoManager::buildGaussianPyramid(const cv::Mat &img, const int levels, std::vector<cv::Mat> &pyramid) {
	if (levels < 1){
		std::cerr << "Levels must be larger than 1" << std::endl;
		return false;
	}

	pyramid.clear();
	cv::Mat currentImg = img;
	for (int i = 0; i<levels; i++) {
		cv::Mat down;
		cv::pyrDown(currentImg, down);
		pyramid.push_back(down);
		currentImg = down;
	}
	return true;
}

/**
* This method implements the upsampling of a gaussian pyramid
*
* @param src the source image
* @param levels levels of the gaussian pyramid
* @param dst the destination image
* @return true if writer could be opened
*/
void VideoManager::upsamplingFromGaussianPyramid(const cv::Mat &src, const int levels, cv::Mat &dst) {
	cv::Mat currentLevel = src.clone();
	for (int i = 0; i < levels; ++i) {
		cv::Mat up;
		cv::pyrUp(currentLevel, up);
		currentLevel = up;
	}
	currentLevel.copyTo(dst);
}

/**
* This method reconstructs an image from given laplacian pyramid
*
* @param pyramid the laplacian pyramid
* @param levels how many levels of the pyramid are to be used
* @param dst resulting image
*/
void VideoManager::reconImgFromLaplacianPyramid(const std::vector<cv::Mat> &pyramid, const int levels, cv::Mat &dst) {
	cv::Mat currentImg = pyramid[levels];
	for (int i = levels - 1; i >= 0; i--) {
		cv::Mat up;
		cv::pyrUp(currentImg, up, pyramid[i].size());
		currentImg = up + pyramid[i];
	}
	dst = currentImg.clone();
}






// #########################################################
// #################### general Getters ####################
// #########################################################
/**
* This method returns whether a video has been openend
*
* @return true if a video has been openend
*/
bool VideoManager::isOpened() {
	return capture.isOpened();
}

/**
* This method returns whether the current video has been stopped
*
* @return true if video has been stopped
*/
bool VideoManager::isStopped() {
	return stopped;
}

/**
* This method returns the codec of the currently open video
*
* @param codec array for the codec
*
* @return an integer identifying the codec
*/
int VideoManager::getCodec(char codec[]) {
	union { int value; char code[4]; } result;		// http://www.willemer.de/informatik/cpp/union.htm

	result.value = static_cast<int>(capture.get(CV_CAP_PROP_FOURCC));

	codec[0] = result.code[0];
	codec[1] = result.code[1];
	codec[2] = result.code[2];
	codec[3] = result.code[3];

	return result.value;
}

/**
* This method returns the framerate of the currently open video
*
* @return framerate of the currently open video
*/
double VideoManager::getFrameRate() {
	//return rate;
	return capture.get(CV_CAP_PROP_FPS);
}

/**
* This method returns the next frame of the current video if there are any
*
* @return the next frame
*/
bool VideoManager::getNextFrame(cv::Mat &frame) {
	return capture.read(frame);
//	if (curPos < length){
//		cv::Mat temp;
//		temp = frames.at(curPos).clone();
//		temp.convertTo(frame, CV_8UC3); //TODO add scaling
//		curPos++;
//		return true;
//	}
//	else return false;
}

/**
* This method returns the size of a frame of the currently open video
*
* @return framesize
*/
cv::Size VideoManager::getFrameSize() {
	int width = static_cast<int>(capture.get(CV_CAP_PROP_FRAME_WIDTH));
	int height = static_cast<int>(capture.get(CV_CAP_PROP_FRAME_HEIGHT));
	return cv::Size(width, height);
}






// ############################################################
// #################### Video Play-Control ####################
// ############################################################
/**
* This method jumps the current video to a given index
*
* @param index the index to be jumped to
*
* @return true if the jump was successfull
*/
bool VideoManager::jumpTo(long index) {
	/*if (index < length){
		curPos = index;
		return true;
	}
	else return false;*/
	
	if (index >= length) {
		return false;		// 1
	}

	cv::Mat frame;
	bool result = capture.set(CV_CAP_PROP_POS_FRAMES, index);

	if (result && !isStopped()){
		capture.read(frame);
		// TODO: show current frame
	}

	return result;
}




void VideoManager::setLevels(int lvls){
	levels = lvls;
}


// ####################################################################
// #################### Color Magnification Setter ####################
// ####################################################################
/**
* This method sets the amplification for color magnification
*
* @param amplification the magnitude of the amplification
*/
void VideoManager::setAlpha(float amplification) {
	alpha = amplification;
}

/**
* This method sets the lower cut-off frequency for color magnification
*
* @param low the lower cut-off frequency, e.g. 0.8Hz (for 50 beats per minute)
*/
void VideoManager::setLowCutoff(float low) {
	fl = low;
}

/**
* This method sets the high cut-off frequency for color magnification
*
* @param high the high cut-off frequency, e.g. 2.0Hz (for 120 beats per minute)
*/
void VideoManager::setHighCutoff(float high) {
	fh = high;
}

/**
* This method sets the low and high cut-off frequencies for color magnification
*
* @param low the lower cut-off frequency, e.g. 0.8Hz (for 50 beats per minute)
* @param high the high cut-off frequency, e.g. 2.0Hz (for 120 beats per minute)
*/
void VideoManager::setFrequencyCutoffs(float low, float high) {
	setLowCutoff(low);
	setHighCutoff(high);
}







// #####################################################################
// #################### Motion Magnification Setter ####################
// #####################################################################

/**
* This method sets the cut-off frequencies for motion magnification
*
* @param lambda cut-off frequency
*/
void VideoManager::setMotionCutOffWavelength(float lambda) {

	lambda_c = lambda;
}

/**
* This method sets the exaggeration factor for motion magnification
*
* @param factor the exaggeration factor
*/
void VideoManager::setExaggerationFactor(float factor) {
	exaggeration_factor = factor;
}

/**
* This method sets the chromatic attenuation for motion magnification
*
* @param attenuation the chromatic attentuation
*/
void VideoManager::setChromAttenuation(float attenuation) {
	chromAttenuation = attenuation;
}