#include "VideoPlayer.h"

VideoPlayer::VideoPlayer(VideoManager* videoManager, QObject *parent)
: vidMgr(videoManager), QThread(parent)
{
	stop = true;
}

VideoPlayer::~VideoPlayer()
{
	mutex.lock();
	stop = true;
	condition.wakeOne();
	mutex.unlock();
	wait();
}

/**
* Sets new input video file in VideoManager and gets the video frame rate.
* @param filename The path to the new input video file.
*
* @return true if loading was successful
*/
bool VideoPlayer::loadVideo(std::string filename) {
	Stop(); //stop playback
	if (vidMgr->setInput(filename)){
		frameRate = vidMgr->getFrameRate();
		return true;
	}
	else {
		return false;
	}
}

/**
* If player is not running, starts the player thread at low priority.
*/
void VideoPlayer::Play()
{
	if (!isRunning()) {
		if (isStopped()){
			stop = false;
		}
		start(LowPriority); //start thread
	}
}

/**
* Gets a frame from the VideoManager, transforms it to a QImage and emits it as processedImage signal. (Overrides the run-method of the thread.)
*/
void VideoPlayer::run()
{
	int delay = (1000 / frameRate); //emit every 1000/framerate millisecs
	while (!stop){
		if (!vidMgr->getNextFrame(frame)) //get frame
		{
			stop = true;
		}

		//convert frame
		cv::cvtColor(frame, RGBframe, CV_BGR2RGB);
		img = QImage((const unsigned char*)(RGBframe.data),
			RGBframe.cols, RGBframe.rows, QImage::Format_RGB888);

		emit processedImage(img); //emit to processedImage signal
		this->msleep(delay); //wait 1000/framerate millisecs
	}
}

/**
* Stops playback.
*/
void VideoPlayer::Stop()
{
	stop = true;
}

/**
* Capsules the sleep-method of QThread.
* @param ms millisecs to sleep
*/
void VideoPlayer::msleep(int ms){
	QThread::msleep(ms);
}

/**
* @return whether playback is stopped
*/
bool VideoPlayer::isStopped() const{
	return this->stop;
}

/**
* @return current frame from the VideoManager
*/
double VideoPlayer::getCurrentFrame(){
	return vidMgr->getCurrentFrame();
}

/**
* @return total number of frames in the video
*/
double VideoPlayer::getNumberOfFrames(){
	return vidMgr->getNumberOfFrames();
}

/**
* @return frame rate of the video
*/
double VideoPlayer::getFrameRate(){
	return frameRate;
}

/**
* Jumps to the specified frame in the VideoManager
* @param frameNumber
*/
void VideoPlayer::setCurrentFrame(int frameNumber)
{
	vidMgr->jumpTo(frameNumber);
}