#include "vidmag.h"
#include <QMouseEvent>

VidMag::VidMag(QWidget *parent)
	: QMainWindow(parent),
	ui(new Ui_VidMagClass)
{
	vidMgr = new VideoManager(); //new manager
	vidPlr = new VideoPlayer(vidMgr); //new player
	QObject::connect(vidPlr, SIGNAL(processedImage(QImage)),
		this, SLOT(updatePlayerUI(QImage))); //connect signal from player to slot in UI; takes frame from VideoManager, processes it and displayes it

	//setMouseTracking(true);
	ui->setupUi(this);

	//disable controls until video is loaded
	ui->button_play->setEnabled(false);
	ui->button_process->setEnabled(false);
	ui->slider_video->setEnabled(false);

	//booleans for masking
	masking = false;
	selectionStarted = false;
}

VidMag::~VidMag()
{
	delete vidPlr;
	delete vidMgr;
	delete ui;
}

/**
* Slot that the signal processedImage connects to. Displays the provided QImage as QPixmap in the video label. Also, updates progress slider and position label.
* @param img frame to display
*/
void VidMag::updatePlayerUI(QImage img)
{
	if (!img.isNull())
	{
		ui->label_video->setAlignment(Qt::AlignCenter); //center video frame
		QPixmap pixmap = QPixmap::fromImage(img).scaled(ui->label_video->size(),
			Qt::KeepAspectRatio, Qt::FastTransformation); //transform to pixmap; scale to size of label
		ui->label_video->setPixmap(pixmap); //set label pixmap

		pixmap_rect = pixmap.rect(); //get dimensions of scaled video frame

		//update slider and position label
		ui->slider_video->setValue(vidPlr->getCurrentFrame());
		ui->label_video_position->setText(getFormattedTime((int)vidPlr->getCurrentFrame() / (int)vidPlr->getFrameRate()));

		if (vidPlr->getCurrentFrame() == vidPlr->getNumberOfFrames() - 1) vidPlr->setCurrentFrame(0); //set playback position to start, when last frame is reached
	}
}

/**
* Opens a file dialog to select a video and tries to load it. If successful, enables the other UI elements and gets video details.
*/
void VidMag::on_button_load_clicked()
{
	//open file dialog
	QString filename = QFileDialog::getOpenFileName(this,
		tr("Open Video"), "../Data/",
		tr("Video Files (*.avi *.mpg *.mp4)"));

	QFileInfo name = filename;

	if (!filename.isEmpty()){ //file name set?
		if (!vidPlr->loadVideo(filename.toLatin1().data())) //can file be loaded?
		{
			QMessageBox msgBox;
			msgBox.setText("The selected video could not be opened!");
			msgBox.exec();
		} else { //get video details and set UI values; start playback
			ui->button_play->setEnabled(true);
			ui->button_process->setEnabled(true);
			ui->slider_video->setEnabled(true);
			ui->slider_video->setMaximum(vidPlr->getNumberOfFrames()-1);
			ui->label_video_length->setText(getFormattedTime((int)vidPlr->getNumberOfFrames() / (int)vidPlr->getFrameRate()));
			on_button_play_clicked();
		}
	}
}

/**
* Starts/stops the video and changes the button text.
*/
void VidMag::on_button_play_clicked()
{
	if (vidPlr->isStopped())
	{
		vidPlr->Play();
		ui->button_play->setText(tr("Stop"));
	}
	else
	{
		vidPlr->Stop();
		ui->button_play->setText(tr("Play"));
	}
}

/**
* Formats seconds to (h:)mm:ss.
* @param timeInSeconds to format
*
* @return string of format h:mm:ss
*/
QString VidMag::getFormattedTime(int timeInSeconds){

	int seconds = (int)(timeInSeconds) % 60;
	int minutes = (int)((timeInSeconds / 60) % 60);
	int hours = (int)((timeInSeconds / (60 * 60)) % 24);

	QTime t(hours, minutes, seconds);
	if (hours == 0)
		return t.toString("mm:ss");
	else
		return t.toString("h:mm:ss");
}

/**
* Stops video playback.
*/
void VidMag::on_slider_video_sliderPressed()
{
	vidPlr->Stop();
	ui->button_play->setText(tr("Play"));
}

/**
* Starts video playback.
*/
void VidMag::on_slider_video_sliderReleased()
{
	vidPlr->Play();
	ui->button_play->setText(tr("Stop"));
}

/**
* Sets the new position in the VideoManager and updates the position label.
* @param position new playback position (as frame number)
*/
void VidMag::on_slider_video_sliderMoved(int position)
{
	vidPlr->setCurrentFrame(position);
	ui->label_video_position->setText(getFormattedTime(position / (int)vidPlr->getFrameRate()));
}

/**
* Sets input values depending on combo box selection.
*/
void VidMag::on_button_presets_load_clicked(){
	switch (ui->val_combo_presets->currentIndex()){
	case 1: {
				ui->val_edit_alpha->setText("50");
				ui->val_edit_chromatic_attenuation->setText("0.1");
				ui->val_edit_cutoff_wavelength->setText("1000");
				ui->val_edit_exaggeration_factor->setText("2");
				ui->val_edit_freqcutoff_high->setText("1");
				ui->val_edit_freqcutoff_low->setText("0.833");
				ui->val_combo_type->setCurrentIndex(0);
				break;
	}
	case 2: {
				ui->val_edit_alpha->setText("10");
				ui->val_edit_chromatic_attenuation->setText("0.1");
				ui->val_edit_cutoff_wavelength->setText("16");
				ui->val_edit_exaggeration_factor->setText("2");
				ui->val_edit_freqcutoff_high->setText("0.4");
				ui->val_edit_freqcutoff_low->setText("0.05");
				ui->val_combo_type->setCurrentIndex(1);
				break;
	}
	case 3: {
				ui->val_edit_alpha->setText("10");
				ui->val_edit_chromatic_attenuation->setText("0.1");
				ui->val_edit_cutoff_wavelength->setText("16");
				ui->val_edit_exaggeration_factor->setText("2");
				ui->val_edit_freqcutoff_high->setText("0.4");
				ui->val_edit_freqcutoff_low->setText("0.05");
				ui->val_combo_type->setCurrentIndex(1);
				break;
	}
	}
}


/**
* Starts/stops masking of a loaded video.
*/
void VidMag::on_button_mask_clicked(){
	if (masking){
		masking = false;
		ui->button_mask->setStyleSheet("Background-color: #ddd; Border: 1px solid #aaa;");
	}
	else {
		masking = true;
		ui->button_mask->setStyleSheet("Background-color: #9f9; Border: 1px solid #aaa;");
	}
	selectionStarted = false;
}

/**
* Removes the masking inputs.
*/
void VidMag::on_button_delete_mask_clicked(){
	masking = false;
	selectionStarted = false;
	ui->val_edit_mask_x1->setText("");
	ui->val_edit_mask_y1->setText("");
	ui->val_edit_mask_x2->setText("");
	ui->val_edit_mask_y2->setText("");
}

/**
* Validates the inputs. If all inputs are set correctly, they are set in the VideoManager, the playback is stopped and the video is processed by the VideoManager.
*/
void VidMag::on_button_process_clicked(){
	bool validInputs = true;

	//read and validate inputs
	float frequencyCutoffLow;
	if (validateLineEdit(ui->val_edit_freqcutoff_low, true)){
		frequencyCutoffLow = ui->val_edit_freqcutoff_low->text().toDouble();
	}
	else validInputs = false;

	float frequencyCutoffHigh;
	if (validateLineEdit(ui->val_edit_freqcutoff_high, true)){
		frequencyCutoffHigh = ui->val_edit_freqcutoff_high->text().toDouble();
	}
	else validInputs = false;

	float alpha;
	if (validateLineEdit(ui->val_edit_alpha, true)){
		alpha = ui->val_edit_alpha->text().toDouble();
	}
	else validInputs = false;

	float cutoffWavelength;
	if (validateLineEdit(ui->val_edit_cutoff_wavelength, true)){
		cutoffWavelength = ui->val_edit_cutoff_wavelength->text().toDouble();
	}
	else validInputs = false;

	float exaggerationFactor;
	if (validateLineEdit(ui->val_edit_exaggeration_factor, true)){
		exaggerationFactor = ui->val_edit_exaggeration_factor->text().toDouble();
	}
	else validInputs = false;

	float chromatic_attenuation;
	if (validateLineEdit(ui->val_edit_chromatic_attenuation, true)){
		chromatic_attenuation = ui->val_edit_chromatic_attenuation->text().toDouble();
	}
	else validInputs = false;

	//masking inputs
	bool masked = true;

	float mask_x1;
	if (validateLineEdit(ui->val_edit_mask_x1, false)){
		mask_x1 = ui->val_edit_mask_x1->text().toDouble();
	}
	else masked = false;

	float mask_y1;
	if (validateLineEdit(ui->val_edit_mask_y1, false)){
		mask_y1 = ui->val_edit_mask_y1->text().toDouble();
	}
	else masked = false;

	float mask_x2;
	if (validateLineEdit(ui->val_edit_mask_x2, false)){
		mask_x2 = ui->val_edit_mask_x2->text().toDouble();
	}
	else masked = false;

	float mask_y2;
	if (validateLineEdit(ui->val_edit_mask_y2, false)){
		mask_y2 = ui->val_edit_mask_y2->text().toDouble();
	}
	else masked = false;


	//if inputs are valid, set parameters in videomanager and start magnification
	if (validInputs){
		//check for switched cutoff values
		if (frequencyCutoffHigh < frequencyCutoffLow){
			float temp = frequencyCutoffHigh;
			frequencyCutoffHigh = frequencyCutoffLow;
			frequencyCutoffLow = temp;
		}

		//set inputs
		vidMgr->setFrequencyCutoffs(frequencyCutoffLow, frequencyCutoffHigh); //temporal
		vidMgr->setAlpha(alpha);
		vidMgr->setMotionCutOffWavelength(cutoffWavelength); //spatial
		vidMgr->setExaggerationFactor(exaggerationFactor);
		vidMgr->setChromAttenuation(chromatic_attenuation);
		if (masked) vidMgr->setMask(mask_x1 / pixmap_rect.width(), mask_y1 / pixmap_rect.height(), mask_x2 / pixmap_rect.width(), mask_y2 / pixmap_rect.height(), true);
		else vidMgr->setMask(-1, -1, -1, -1, false);

		//stop video, if still playing
		if (!vidPlr->isStopped()) on_button_play_clicked();
		//change video frame to "processing..."
		cv::Mat processingImage(360, 640, CV_8UC3, cv::Scalar(0));
		cv::putText(processingImage, "processing...", cv::Point2d(260, 180), FONT_HERSHEY_COMPLEX_SMALL, 0.8, cv::Scalar(255, 255, 255), 1, CV_AA);
		QImage img = QImage((const unsigned char*)(processingImage.data), processingImage.cols, processingImage.rows, QImage::Format_RGB888);
		ui->label_video->setAlignment(Qt::AlignCenter);
		ui->label_video->setPixmap(QPixmap::fromImage(img).scaled(ui->label_video->size(),
			Qt::KeepAspectRatio, Qt::FastTransformation));
		//disable inputs
		ui->centralWidget->setEnabled(false);
		//repaint ui
		repaint();

		//magnify color/motion
		switch (ui->val_combo_type->currentIndex()){
			case 0: {
						vidMgr->colorMagnification();
						break;
			}
			case 1: {
						vidMgr->motionMagnification();
						break;
			}
		}

		//play magnified video
		if (vidPlr->isStopped()) on_button_play_clicked();

		//enable inputs
		ui->centralWidget->setEnabled(true);
	}
}

/**
* Validates the text of a line edit. It has to be a non-negative double or integer value.
* @param lineEdit UI element to validate
* @param format true, if the style of the UI element should be altered depending on the validity
*
* @return true if the text of the provided line edit is valid
*/
bool VidMag::validateLineEdit(QLineEdit* lineEdit, bool format){
	QString text = lineEdit->text();
	bool* ok = new bool;
	text.toDouble(ok);
	
	if (*ok && text.toDouble() >= 0.0){
		if(format) lineEdit->setStyleSheet("border: 1px solid #666666;");
		return true;
	}
	else {
		if(format) lineEdit->setStyleSheet("border: 1px solid #ff0000;");
		return false;
	}
}

/**
* If state is set to masking, start a new selection. A QRubberBand is displayed where the selection intersects the video frame.
* @param e mouse event to process
*/
void VidMag::mousePressEvent(QMouseEvent *e)
{
	if (masking){
		QPoint mouse = e->pos();
		if (!pixmap_rect.contains(pointToPixmap(mouse))){ //mouse position not inside video frame -> set to position inside video frame
			QPoint mousePixmap = pointToPixmap(mouse);
			if (mousePixmap.x() <= 0){
				mouse.setX(pixmapTl().x());
			}
			else if (mousePixmap.x() > pixmap_rect.width()){
				mouse.setX(pixmapTl().x() + pixmap_rect.width());
			}

			if (mousePixmap.y() <= 0){
				mouse.setY(pixmapTl().y());
			}
			else if (mousePixmap.y() > pixmap_rect.height()){
				mouse.setY(pixmapTl().y() + pixmap_rect.height());
			}
		}
		selectionStarted = true;
		//init selection rectangle
		selectionRect.setTopLeft(mouse);
		selectionRect.setBottomRight(mouse);

		//visuzalize selection
		selection = new QRubberBand(QRubberBand::Rectangle, this);
		selection->setGeometry(selectionRect);
		selection->show();
	}
}

/**
* If a selection has been started, the mouse position updates the selection rectangle.
* @param e mouse event to process
*/
void VidMag::mouseMoveEvent(QMouseEvent *e)
{
	if (selectionStarted)
	{
		QPoint mouse = e->pos();
		if (!pixmap_rect.contains(pointToPixmap(mouse))){ //mouse position not inside video frame -> set to position inside video frame
			QPoint mousePixmap = pointToPixmap(mouse);
			if (mousePixmap.x() <= 0){
				mouse.setX(pixmapTl().x());
			}
			else if (mousePixmap.x() > pixmap_rect.width()){
				mouse.setX(pixmapTl().x() + pixmap_rect.width());
			}

			if (mousePixmap.y() <= 0){
				mouse.setY(pixmapTl().y());
			}
			else if (mousePixmap.y() > pixmap_rect.height()){
				mouse.setY(pixmapTl().y() + pixmap_rect.height());
			}
		}
		//update selection rectangle
		selectionRect.setBottomRight(mouse);
		selection->setGeometry(selectionRect.normalized()); //flip rectangle corners, if selection direction is inverted
	}
}

/**
* If a selection has been started, the selection rectangle is transformed to coordinates relative to the video frame. 
* The x/y-coordinates of the top-left and the bottom-right point of the selection rectangle are set as text of line edits in the UI and the QRubberBand is hid.
* @param e mouse event to process
*/
void VidMag::mouseReleaseEvent(QMouseEvent *e)
{
	if (selectionStarted){
		//transform corner point positions and set them in the line edits
		QPoint tl = ui->label_video->pos();
		selectionRect.setTopLeft(selectionRect.topLeft() - tl);
		selectionRect.setBottomRight(selectionRect.bottomRight() - tl);

		QRect label = ui->label_video->rect();
		int xOffset = (label.width() - pixmap_rect.width()) / 2;
		int yOffset = (label.height() - pixmap_rect.height()) / 2;

		int* x1 = new int;
		int* y1 = new int;
		int* x2 = new int;
		int* y2 = new int;
		selectionRect.normalized().getCoords(x1, y1, x2, y2);

		ui->val_edit_mask_x1->setText(QString::number(*x1 - xOffset));
		ui->val_edit_mask_y1->setText(QString::number(*y1 - yOffset));
		ui->val_edit_mask_x2->setText(QString::number(*x2 - xOffset));
		ui->val_edit_mask_y2->setText(QString::number(*y2 - yOffset));

		//end selection and hide QRubberBand
		selectionStarted = false;
		masking = false;
		ui->button_mask->setStyleSheet("Background-color: #ddd; Border: 1px solid #aaa;");
		selection->hide();
	}
}


/**
* Transforms a point from coordinates relative to the window to coordinates relative to the video frame.
* @param point to transform
*
* @return point with coordinates relative to the video frame
*/
QPoint VidMag::pointToPixmap(QPoint point){
	QRect label = ui->label_video->rect();
	int xOffset = (label.width() - pixmap_rect.width()) / 2;
	int yOffset = (label.height() - pixmap_rect.height()) / 2;
	QPoint offset = QPoint(xOffset, yOffset);

	return point - ui->label_video->pos() - offset;
}

/**
* @return top-left point of the video frame in window coordinates
*/
QPoint VidMag::pixmapTl(){
	QRect label = ui->label_video->rect();
	int xOffset = (label.width() - pixmap_rect.width()) / 2;
	int yOffset = (label.height() - pixmap_rect.height()) / 2;
	QPoint offset = QPoint(xOffset, yOffset);

	return ui->label_video->pos() + offset;
}